//
//  LIVOpenGLView.h
//  Large Image Viewer
//
//  Created by Tom Bernard on 10/11/06.
//  Copyright Bersearch Information Services 2006 . All rights reserved.
//
/* LIVOpenGLView */

#import <Cocoa/Cocoa.h>
#import <OpenGL/gl.h>
#import <OpenGL/glext.h>

/*
typedef struct _TileImageData
{
	char *tileImagePixels;
	unsigned int tipIndex;
		
} TileImageData;

typedef TileImageData *TileImageDataPointer;
 */


@class Tile;
@class LIVDocument;
@interface LIVOpenGLView : NSOpenGLView
{
	IBOutlet LIVDocument		*document;
	GLdouble					orthoDelta, oldOrthoDelta, orthoDeltaDelta, orthoDeltaScaleFactor;
	GLdouble					orthoXstart, orthoXend, orthoYstart, orthoYend, orthoZstart, orthoZend;
	long						swap_enabled;
	BOOL						inited, outputOpenBenchmarks, saveTileImagesToTIFF, drawTileOutlines;
	GLuint						imageTexture;
	int							bitsPerPixel, bytesPerRow, bytesPerPlane, numberOfPlanes;
	NSString					*colorspaceName;
	NSBitmapFormat				bitmapFormat;
	int							bitsPerSample, bytesPerSample, samplesPerPixel, imagePixelsWide, imagePixelsHigh;
	int							calculatedBytesPerRow, pad;
	unsigned int				tileRows, tileColumns, tileWidth, tileHeight, lastRowHeight, lastColumnWidth;
	unsigned int				textureWidth, textureHeight, lastColumnTextureWidth, lastRowTextureHeight;
	unsigned char				*imageBitmapData;
	BOOL						imageHasAlpha, loadSucceeded;
	BOOL						imageXNeedsFlip, imageYNeedsFlip, imageZNeedsFlip;
	BOOL						tilesNeedFlip;
	BOOL						optionKeyDown, controlKeyDown, commandKeyDown, shiftDown, capsLockDown;
	NSMutableArray				*tileArray;
	NSString					*fileNameForDisplay, *filePath, *diagnosticFolderPath, *livLogFilePath;
	float						imageAspectRatio, windowDeltaWidth, windowDeltaHeight;
	UnsignedWide				startBenchmarkUW;
	NSDateFormatter				*dateFormatter;
}
#pragma mark initialization
- (void)delayedInit;
- (NSRect)frameForImageWithWidth:(float)width height:(float)height;
- (BOOL)loadTilesFromImageRep:(NSBitmapImageRep *)imageRep;
- (BOOL)doTileLoad;
- (BOOL)saveTileImageWithPixels:(char *)pixels
					 pixelsWide:(int)width 
					 pixelsHigh:(int)height 
				  bitsPerSample:(int)bps 
				samplesPerPixel:(int)spp 
					   hasAlpha:(BOOL)alpha 
						 toFile:(NSString *)path;

- (BOOL)saveTileImageWithPixels:(char **)pixels
					 pixelsWide:(int)width 
					 pixelsHigh:(int)height 
				  bitsPerSample:(int)bps 
				samplesPerPixel:(int)spp 
					   hasAlpha:(BOOL)alpha 
					   isPlanar:(BOOL)isPlanar 
				 colorSpaceName:(NSString *)colorSpaceName 
					bytesPerRow:(int)rowBytes 
				   bitsPerPixel:(int)pixelBits
						 toFile:(NSString *)path;

- (unsigned int)tileIndexForRow:(unsigned int)_row column:(unsigned int)_column;
- (float)imageAspectRatio;
- (BOOL)glPredatesGLOnePointOne;
#pragma mark drawing
- (void)drawRect:(NSRect)rect;
- (BOOL)shouldDrawTile:(Tile *)tile;
#pragma mark keyboard response
- (void)clearFlags;
- (void)setModifiersFromCarbon;
- (void)setFlagsfromEvent:(NSEvent *)theEvent;
#pragma mark zoom and translation
- (void)zoomIn;
- (void)zoomOut;
- (void)finishZoom;
- (void)translateForLeftArrowDown;
- (void)translateForRightArrowDown;
- (void)translateForUpArrowDown;
- (void)translateForDownArrowDown;
- (void)translateRightByFraction:(GLdouble)fraction;
- (void)translateLeftByFraction:(GLdouble)fraction;
- (void)translateUpByFraction:(GLdouble)fraction;
- (void)translateDownByFraction:(GLdouble)fraction;
- (void)translateImageLeftEdgeToViewLeftEdge;
- (void)translateImageRightEdgeToViewRighgtEdge;
- (void)translateImageTopEdgeToViewTopEdge;
- (void)translateImageBottomEdgeToViewBottomEdge;
#pragma mark diagnostic
- (void)toggleDrawTileOutlines;
- (void)startBenchmark;
- (void)outputBenchmarkWithMessage:(NSString *)message;
- (void)outputFileInfo;
- (NSString *)diagnosticFolderPathForComponent:(NSString *)_component;
- (NSString *)folderForTIFFsFromFile:(NSString *)_fileNameForDisplay;
- (NSString *)dateTimeStampAsString;
- (NSDateFormatter *)dateFormatter;
- (NSString *)determineLIVLogFlePath;
- (void)logMessageToLIVLog:(NSString *)msg;
- (void)dateTimeStampToLog;
#pragma mark accessors
- (NSString *)colorspaceName;
- (void)setColorspaceName:(NSString *)_colorspaceName;
- (NSMutableArray *)tileArray;
- (void)setTileArray:(NSMutableArray *)_tileArray;
- (NSString *)fileNameForDisplay;
- (void)setFileNameForDisplay:(NSString *)_fileNameForDisplay;
- (NSString *)filePath;
- (void)setFilePath:(NSString *)_filePath;
- (NSString *)diagnosticFolderPath;
- (void)setDiagnosticFolderPath:(NSString *)_diagnosticFolderPath;
- (NSString *)livLogFilePath;
- (void)setLivLogFilePath:(NSString *)_livLogFilePath;
@end

