//
//  LIVOpenGLView.m
//  Large Image Viewer
//
//  Created by Tom Bernard on 10/11/06.
//  Copyright Bersearch Information Services 2006 . All rights reserved.
//

#import "LIVOpenGLView.h"
#import "Tile.h"
#import "LIVDocument.h"
#import <Carbon/Carbon.h>

NSString *livDiagnosticFolder =		@"LIV Diagnostic Folder";
NSString *livLog =					@"LIV Log.txt";
NSString *nlStr =					@"\n";


void printGLError(char *label, GLenum error)
{
	char *errStr;
	
	switch (error)
	{
		case GL_NO_ERROR:
			errStr = "GL_NO_ERROR";
			break;
			
		case GL_INVALID_ENUM:
			errStr = "GL_INVALID_ENUM";
			break;
			
		case GL_INVALID_VALUE:
			errStr = "GL_INVALID_VALUE";
			break;
			
		case GL_INVALID_OPERATION:
			errStr = "GL_INVALID_OPERATION";
			break;
			
		case GL_STACK_OVERFLOW:
			errStr = "GL_STACK_OVERFLOW";
			break;
			
		case GL_STACK_UNDERFLOW:
			errStr = "GL_STACK_UNDERFLOW";
			break;
			
		case GL_OUT_OF_MEMORY:
			errStr = "GL_OUT_OF_MEMORY";
			break;
			
		default: errStr = "undefined error";
	}
	printf("%s%s\n", label, errStr);
}


/*
 
BOOL isDirectoryAtPath(NSString *_path)

 Returns YES if a directory is at _path and the directory is not a package.
 
 */


BOOL isDirectoryAtPath(NSString *_path)
{
	BOOL isDir;
	
	if ([[NSFileManager defaultManager] fileExistsAtPath:_path isDirectory:&isDir])
	{
		NSWorkspace *workspace = [NSWorkspace sharedWorkspace];
		if (isDir && ![workspace isFilePackageAtPath:_path]) return YES;
	}
	return NO;
}



@implementation LIVOpenGLView


#pragma mark initialization


- (id)initWithFrame:(NSRect)frameRect
		pixelFormat:(NSOpenGLPixelFormat*)format
{	
    self = [super initWithFrame:frameRect pixelFormat:format];
	if (self)
	{
		orthoDelta = 1.0;
		orthoDeltaDelta = 0.05;
		orthoDeltaScaleFactor = 0.95;
		
		orthoXstart		= 0.0; 
		orthoXend		= orthoXstart + orthoDelta; 
		orthoYstart		= 0.0; 
		orthoYend		= orthoYstart + orthoDelta; 
		orthoZstart		= 0.0; 
		orthoZend		= 1.0;
		
		imageXNeedsFlip = NO;
		imageYNeedsFlip = YES;
		imageZNeedsFlip = NO;
		
		tilesNeedFlip = NO;
		
		// diagnostics
		
		[self setDiagnosticFolderPath:[self diagnosticFolderPathForComponent:livDiagnosticFolder]];
		[self setLivLogFilePath:[self determineLIVLogFlePath]];
		
		[self setModifiersFromCarbon];
		if (optionKeyDown)
		{
			[self dateTimeStampToLog];
			 outputOpenBenchmarks = YES;
			[self startBenchmark];
		}
		if (commandKeyDown) saveTileImagesToTIFF = YES;
				
	}
	return self;
}




- (id)initWithFrame:(NSRect)frameRect
{
	// printf("-[%s initWithFrame:]\n", [[self description] UTF8String]);
    NSOpenGLPixelFormatAttribute attrs[] =
    {
        NSOpenGLPFADoubleBuffer,
        NSOpenGLPFADepthSize, (NSOpenGLPixelFormatAttribute)32,
        (NSOpenGLPixelFormatAttribute)nil
    };
    
    NSOpenGLPixelFormat* pixFmt = [[[NSOpenGLPixelFormat alloc] initWithAttributes:attrs] autorelease];
    
	
	return [self initWithFrame:frameRect pixelFormat:pixFmt];
}

/*
 */
- (void)dealloc
{
	[self setColorspaceName:nil];
	[self setTileArray:nil];
	[self setFileNameForDisplay:nil];
	[self setFilePath:nil];
	[self setDiagnosticFolderPath:nil];
	[self setLivLogFilePath:nil];
	[dateFormatter release];

	[super dealloc];
}

- (void)delayedInit
{
	if(!inited)
	{
        [[self openGLContext] makeCurrentContext];
        
		swap_enabled = 1;
		[[self openGLContext] setValues:&swap_enabled forParameter:NSOpenGLCPSwapInterval];
		
		glEnable (GL_TEXTURE_2D);
		glPixelStorei(GL_UNPACK_ROW_LENGTH, textureWidth);		// not needed?
		glPixelStorei(GL_UNPACK_ALIGNMENT, 2);
		
		glClearColor(0.72, 0.72, 0.72, 1.0);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
        
        inited = YES;
		
		// set imageAspectRatio to default window content view aspect ratio
		
		NSRect windowFrame = [[self window] frame];
		NSRect myFrame = [self frame];
		windowDeltaWidth = windowFrame.size.width - myFrame.size.width;
		windowDeltaHeight = windowFrame.size.height - myFrame.size.height;
		imageAspectRatio = myFrame.size.height / myFrame.size.width;
		
		
		// set the file name
		
		NSURL *fileURL = [document fileURL];
		[self setFilePath:[fileURL path]];
		[self setFileNameForDisplay:[filePath lastPathComponent]];
		
		
		// get the image from the document, tile it and release it
		
		NSBitmapImageRep *imageRep = [document imageRep];
		loadSucceeded = [self loadTilesFromImageRep:imageRep];
		[document setImageRep:nil];
		
		
		// set the window title
		
		NSString *windowTitle;
		if (loadSucceeded)
		{
			windowTitle = [NSString stringWithFormat:@"%@ %d x %d pixels", fileNameForDisplay, imagePixelsWide, imagePixelsHigh];
		}
		else
		{
			windowTitle = @"Open image failed.";
		}
		
		[[self window] setTitle:windowTitle];
		// [[self window] setNeedsDisplay:YES];
		
		// adjust window frame for image
		
		if (loadSucceeded)
		{
			NSRect windowFrame = [self frameForImageWithWidth:(float)imagePixelsWide height:(float)imagePixelsHigh];
			[[self window] setFrame:windowFrame display:YES];
		}
		
		glMatrixMode(GL_MODELVIEW);
    }
}


- (NSRect)frameForImageWithWidth:(float)width height:(float)height
{
	NSRect frameForImage;
	NSRect desiredContentRect;
	NSRect visibleFrame = [[NSScreen mainScreen] visibleFrame];
	imageAspectRatio = height / width;
	
	if ( (width + windowDeltaWidth) > visibleFrame.size.width && (height + windowDeltaHeight) > visibleFrame.size.height) 
	{
		// image with window frame exceeds visible rect in both directions
		// use the relative aspect ratios to decide which dimension governs
		
		float var = visibleFrame.size.height / visibleFrame.size.width;
		if (var > imageAspectRatio)
		{
			desiredContentRect = NSMakeRect(0.0, 0.0, visibleFrame.size.width - windowDeltaWidth, 0.0);
			desiredContentRect.size.height = visibleFrame.size.width * imageAspectRatio;
		}
		else 
		{
			desiredContentRect = NSMakeRect(0.0, 0.0, 0.0, visibleFrame.size.height - windowDeltaHeight);
			desiredContentRect.size.width = visibleFrame.size.height / imageAspectRatio;
		}

	}
	else
		if ( (width + windowDeltaWidth) > visibleFrame.size.width || (height + windowDeltaHeight) > visibleFrame.size.height)
		{
			// image with window frame exceeds visible rect in one direction
			// return largest frame that fits within visible rect
			
			if ((width + windowDeltaWidth) > visibleFrame.size.width)
			{
				// image is wider than screen width
				// use visible frame width and calculate height
				
				desiredContentRect = NSMakeRect(0.0, 0.0, visibleFrame.size.width - windowDeltaWidth, 0.0);
				desiredContentRect.size.height = visibleFrame.size.width * imageAspectRatio;
			}
			else
			{
				// image is taller than screen height
				// use visible frame height and calculate width
				
				desiredContentRect = NSMakeRect(0.0, 0.0, 0.0, visibleFrame.size.height - windowDeltaHeight);
				desiredContentRect.size.width = visibleFrame.size.height / imageAspectRatio;
			}
			
		}
		else 
		{
			// image and window frame will fit within visible rect
			// return frame that centers image within visible rect
			
			desiredContentRect = NSMakeRect(0.0, 0.0, width, height);
		}
	
	frameForImage = [[self window] frameRectForContentRect:desiredContentRect];
	frameForImage.origin.x = visibleFrame.origin.x + (visibleFrame.size.width - frameForImage.size.width) / 2.0;
	frameForImage.origin.y = visibleFrame.origin.y + (visibleFrame.size.height - frameForImage.size.height) / 2.0;
	return frameForImage;
}



/*
// NSEPSImageRep NSTIFFImageRep
- (NSString *)fileNameOfFileToOpenWithTypes:(NSArray *)_types
{
	int button = NSCancelButton;
	NSOpenPanel *openPanel = [NSOpenPanel openPanel];
	button = [openPanel runModalForTypes:_types];
	[self setFlagsfromEvent:[openPanel currentEvent]];
	if (button == NSOKButton)
	{
		return [[openPanel filenames] objectAtIndex:0];
	}
	else
	{
		return nil;
	}
}
 */






- (void) reshape
{
    if(inited)
    {
		[[self openGLContext] makeCurrentContext];
		
		NSRect frame = [self frame];
		glViewport(0, 0, frame.size.width, frame.size.height);
		
		[self setNeedsDisplay:YES];
    }
}




- (BOOL)loadTilesFromImageRep:(NSBitmapImageRep *)imageRep
{
	loadSucceeded = NO;
	
	if (outputOpenBenchmarks) [self outputBenchmarkWithMessage:@"Begin -loadTilesFromImageRep:\n"];
	
	if (!imageRep)
	{
		printf("No imageRep from data\n");
		return loadSucceeded;
	}
	
	// colorspaceName = [imageRep colorSpaceName];
	[self setColorspaceName:[imageRep colorSpaceName]];
	bitmapFormat = [imageRep bitmapFormat];
	bitsPerSample = [imageRep bitsPerSample];
	bytesPerSample = bitsPerSample / 8;
	samplesPerPixel = [imageRep samplesPerPixel];
	bitsPerPixel = [imageRep bitsPerPixel];
	bytesPerPlane = [imageRep bytesPerPlane];
	numberOfPlanes = [imageRep numberOfPlanes];
	imagePixelsWide = [imageRep pixelsWide];
	imagePixelsHigh = [imageRep pixelsHigh];
	bytesPerRow = [imageRep bytesPerRow];
	calculatedBytesPerRow = bitsPerSample * samplesPerPixel * imagePixelsWide / 8;
	pad = bytesPerRow - calculatedBytesPerRow;
	imageHasAlpha = [imageRep hasAlpha];
	imageBitmapData = [imageRep bitmapData];
	
	
	
	// GL_MAX_TEXTURE_SIZE
	
	GLint max_texture_size;
	glGetIntegerv(GL_MAX_TEXTURE_SIZE, &max_texture_size);
	
	// per discussion with David Duncan, Apple DTS, smaller textures based on screen size are better
	// max_texture_size = 1024; stopped most Leopard tile blackout blues
	
	max_texture_size = 512;
	
	// widths and heights are in pixels
	
	if (imagePixelsWide > max_texture_size)
	{
		tileColumns = (int)ceil((float)imagePixelsWide / max_texture_size);
		tileWidth = max_texture_size;
		textureWidth = max_texture_size;
		lastColumnWidth = fmod(imagePixelsWide, max_texture_size);
		lastColumnWidth = lastColumnWidth ? lastColumnWidth : max_texture_size;
		lastColumnTextureWidth = 64;
		while (lastColumnTextureWidth < lastColumnWidth) lastColumnTextureWidth *= 2;
	}
	else
	{
		tileColumns = 1;
		tileWidth = 0;
		textureWidth = 0;
		lastColumnWidth = imagePixelsWide;
		lastColumnTextureWidth = 64;
		while (lastColumnTextureWidth < lastColumnWidth) lastColumnTextureWidth *= 2;
	}
	
	if (imagePixelsHigh > max_texture_size)
	{
		tileRows = (int)ceil((float)imagePixelsHigh / max_texture_size);
		tileHeight = max_texture_size;
		textureHeight = max_texture_size;
		lastRowHeight = fmod(imagePixelsHigh, max_texture_size);
		lastRowHeight = lastRowHeight ? lastRowHeight : max_texture_size;
		lastRowTextureHeight = 64;
		while (lastRowTextureHeight < lastRowHeight) lastRowTextureHeight *= 2;
	}
	else
	{
		tileRows = 1;
		tileHeight = 0;
		textureHeight = 0;
		lastRowHeight = imagePixelsHigh;
		lastRowTextureHeight = 64;
		while (lastRowTextureHeight < lastRowHeight) lastRowTextureHeight *= 2;
	}
	
	
	loadSucceeded = [self doTileLoad];
	
	// [image release];
	// image = nil;
	
	/*
	 */
	
	
	/*
	[imageRep release];
	imageRep = nil;
	 */
	
	/*
	// set imageAspectRatio to default window aspect ratio
	
	NSRect windowFrame = [[self window] frame];
	imageAspectRatio = windowFrame.size.height / windowFrame.size.width;
	 */
	
	/*
	if (loadSucceeded)
	{
		// if loadSucceeded adjust window size to aspect ratio of image and update imageAspectRatio
		
		NSRect windowFrame = [[self window] frame];
		NSRect myFrame = [self frame];
		NSSize adjustedSize = myFrame.size;
		imageAspectRatio = (float)imagePixelsHigh / imagePixelsWide;
		if (imagePixelsHigh > imagePixelsWide)
		{
			adjustedSize.width = myFrame.size.height / imageAspectRatio;
		}
		else
		{
			adjustedSize.height = myFrame.size.width * imageAspectRatio;
		}
		adjustedSize.width += windowDeltaWidth;
		adjustedSize.height += windowDeltaHeight;
		windowFrame.size = adjustedSize;
		[[self window] setFrame:windowFrame display:YES];
		// [[self window] orderFront:nil];
	}
	 */
	
	
	return loadSucceeded;
}


/*
- (void)doTileLoad

Normally called from -loadTiles
Assumes instance variables 
	
	tileRows, 
	tileColumns, 
	tileWidth, 
	lastColumnWidth, 
	tileHeight, 
	lastRowHeight
	
aready set

	textureCount = tileRows * tileColumns tiles
allocates memory for textureCount bit images to be read into textureCount textures

reads imageBitmapData, tiling data into target texture bit images

generates 

tileRows texture rows by tileColumns texture columns

 */

- (BOOL)doTileLoad
{
	// BOOL loadSucceeded = NO;
	
	if (outputOpenBenchmarks) [self outputBenchmarkWithMessage:@"set up for tiling; allocate tile memory\n"];
	
	unsigned int  tileRow, tileColumn, t, textureCount, textureSize;
	// unsigned int row, column, rowLimit, columnLimit;
	
	textureCount = tileRows * tileColumns;
	// textureSize = textureWidth * textureHeight;
	
	int lastTileRow = tileRows - 1;
	int lastTileColumn = tileColumns - 1;
	
	// allocate space for tilesPixelsArray
	
	void *tilesPixelsArray = calloc(textureCount, sizeof(unsigned int *));
	if (!tilesPixelsArray)
	{
		printf("tilesPixelsArray allocation failed.\n");
		return loadSucceeded;
	}
	
	// initialize tilesPixelsArray
	
	for (tileRow = 0; tileRow < tileRows; tileRow++)
	{
		for (tileColumn = 0; tileColumn < tileColumns; tileColumn++)
		{
			t = [self tileIndexForRow:tileRow column: tileColumn];
			textureSize = ((tileColumn == lastTileColumn) ? lastColumnTextureWidth : textureWidth) * 
				((tileRow == lastTileRow) ? lastRowTextureHeight : textureHeight);								// texels
			// ((unsigned int *)tilesPixelsArray)[t] = (unsigned int)calloc(textureSize, bytesPerSample * (imageHasAlpha ? 4 : 3));
			((unsigned int *)tilesPixelsArray)[t] = (unsigned int)calloc(textureSize, bytesPerSample * samplesPerPixel);
			if (!((unsigned int *)tilesPixelsArray)[t])
			{
				printf("texture bit image allocation failed.\n");
				printf("t = %d, textureSize = %d, bytes per texel = %d\n", t, textureSize, imageHasAlpha ? 4 : 3);
				return loadSucceeded;
			}
		}
	}
	
	// initialize workingImageBitmapData
	
	unsigned char *workingImageBitmapData = imageBitmapData;
	
	// allocate space for workingTilesPixelsArray
	
	void *workingTilesPixelsArray = calloc(textureCount, sizeof(unsigned int *));
	if (!workingTilesPixelsArray)
	{
		printf("workingTilesPixelsArray allocation failed.\n");
		return loadSucceeded;
	}
	
	// initialize workingTilesPixelsArray
	
	memcpy(workingTilesPixelsArray, tilesPixelsArray, textureCount * sizeof(unsigned int *));
	
	if (outputOpenBenchmarks) [self outputBenchmarkWithMessage:@"tile the image pixels to tile images\n"];
	
	
	// copy the pixels
	
	for (tileRow = 0; tileRow < tileRows; tileRow++)
	{
		unsigned int tilePixelRow, tilePixelRowsThisTile;
		tilePixelRowsThisTile = (tileRow == lastTileRow) ? lastRowHeight : tileHeight;
		
		for (tilePixelRow = 0; tilePixelRow < tilePixelRowsThisTile; tilePixelRow++)
		{
			unsigned int tileColumn;
			
			for (tileColumn = 0; tileColumn < tileColumns; tileColumn++)
			{
				unsigned int columnWidthThisTile;
				columnWidthThisTile = (tileColumn == lastTileColumn) ? lastColumnWidth : tileWidth;
				t = [self tileIndexForRow:tileRow column: tileColumn];
				unsigned int columnPixelIndex = tileColumn * tileWidth * samplesPerPixel;
				char *sourcePixels = (char *)((unsigned int)workingImageBitmapData + columnPixelIndex);
				char *destPixels = (char *)((unsigned int *)workingTilesPixelsArray)[t];
				unsigned int bytesToCopy = columnWidthThisTile * samplesPerPixel * bytesPerSample;
				
				memcpy(destPixels, sourcePixels, bytesToCopy);
				
				((unsigned int *)workingTilesPixelsArray)[t] += ((tileColumn == lastTileColumn) ? lastColumnTextureWidth : textureWidth) * bytesPerSample * samplesPerPixel;
			}
			workingImageBitmapData += bytesPerRow;
			
		}
	}
	
	
	
	if (saveTileImagesToTIFF)
	{
		// diagnostic
		// save the tile images to TIFF files
		
		if (outputOpenBenchmarks) [self outputBenchmarkWithMessage:@"save tile images to tiff files\n"];
		
		NSString *folderForTIFFs = [self folderForTIFFsFromFile:fileNameForDisplay];
		
		for (tileRow = 0; tileRow < tileRows; tileRow++)
		{
			for (tileColumn = 0; tileColumn < tileColumns; tileColumn++)
			{
				t = [self tileIndexForRow:tileRow column: tileColumn];
				
				NSString *filePathForTIFF = [folderForTIFFs stringByAppendingPathComponent:[NSString stringWithFormat:@"Row %d Column %d Tile %d.tiff", tileRow, tileColumn, t]];
				[self saveTileImageWithPixels:(char *)((unsigned int *)tilesPixelsArray)[t]
								   pixelsWide:(tileColumn == lastTileColumn) ? lastColumnTextureWidth : textureWidth
								   pixelsHigh:(tileRow == lastTileRow) ? lastRowTextureHeight : textureHeight
								bitsPerSample:bitsPerSample
							  samplesPerPixel:samplesPerPixel
									 hasAlpha:imageHasAlpha
									   toFile:filePathForTIFF];
			}
		}
		
	}
	
	
	// debugging
	// return loadSucceeded;
	
	
	
	
	
	// allocate and initialize the tiles
	
	if (outputOpenBenchmarks) [self outputBenchmarkWithMessage:@"alloc init tile objects and textures\n"];
	
	[self setTileArray:[NSMutableArray arrayWithCapacity:0]];
	
	for (tileRow = 0; tileRow < tileRows; tileRow++)
	{
		for (tileColumn = 0; tileColumn < tileColumns; tileColumn++)
		{
			t = [self tileIndexForRow:tileRow column: tileColumn];
			
			Tile *tile = [[[Tile alloc] initWithTexelWidth:(tileColumn == lastTileColumn) ? lastColumnTextureWidth : textureWidth
											   texelHeight:(tileRow == lastTileRow) ? lastRowTextureHeight : textureHeight
													   row:tileRow
													column:tileColumn
											 tileNeedsFlip:tilesNeedFlip
											 colorspaeName:colorspaceName
											 imageHasAlpha:imageHasAlpha
											 bitsPerSample:bitsPerSample
													pixels:(char *)((unsigned int *)tilesPixelsArray)[t] ] autorelease];
			
			free((char *)((unsigned int *)tilesPixelsArray)[t]);
			((unsigned int *)tilesPixelsArray)[t] = 0;
			
			if (tile)
			{
				
				// calculate quad vertices
				unsigned int tilePixelRowsThisTile, columnWidthThisTile;
				tilePixelRowsThisTile = (tileRow == lastTileRow) ? lastRowHeight : tileHeight;
				columnWidthThisTile = (tileColumn == lastTileColumn) ? lastColumnWidth : tileWidth;
				
				GLfloat xl, xr, yb, yt;
				
				xl = (float)tileColumn*textureWidth/(float)imagePixelsWide;
				xr = ((float)tileColumn*textureWidth + (float)((tileColumn == lastTileColumn) ? lastColumnWidth : textureWidth))/imagePixelsWide;
				
				yb = (float)tileRow*textureHeight/(float)imagePixelsHigh;
				yt = ((float)tileRow*textureHeight + (float)((tileRow == lastTileRow) ? lastRowHeight : textureHeight))/imagePixelsHigh;
				
				[tile setVerticeX1:xl
								y1:yb
					
								x2:xr
								y2:yb
					
								x3:xr 
								y3:yt
					
								x4:xl
								y4:yt];
				
				// calculate texture coordinates
				
				GLfloat sl, sr, tb, tt;
				
				sl = 0.0;
				sr = (tileColumn == lastTileColumn) ? (float)lastColumnWidth / (float)lastColumnTextureWidth : 1.0;
				
				tb = 0.0;
				tt = (tileRow == lastTileRow) ? (float)lastRowHeight / (float)lastRowTextureHeight : 1.0;
				
				[tile setTextureCoordinateS1:sl
										  t1:tb
					
										  s2:sr
										  t2:tb
					
										  s3:sr
										  t3:tt
					
										  s4:sl
										  t4:tt];
				
				// add an initialized and located tile to the tile array
				
				[tileArray addObject:tile];
			}
			else
			{
				printf("Tile allocation failed.\n");
				printf("t = %d, tileRow = %d, tileColumn = %d\n", t, tileRow, tileColumn);
				return loadSucceeded;
				
			}
		}
	}
	
	
	if (outputOpenBenchmarks) [self outputBenchmarkWithMessage:@"load complete\n"];
	
	loadSucceeded = YES;
	return loadSucceeded;
}



- (BOOL)saveTileImageWithPixels:(char *)pixels
					 pixelsWide:(int)width 
					 pixelsHigh:(int)height 
				  bitsPerSample:(int)bps 
				samplesPerPixel:(int)spp 
					   hasAlpha:(BOOL)alpha 
						 toFile:(NSString *)path
{
	return [self saveTileImageWithPixels:&pixels
							  pixelsWide:width 
							  pixelsHigh:height 
						   bitsPerSample:bps 
						 samplesPerPixel:spp 
								hasAlpha:alpha 
								isPlanar:NO 
						  colorSpaceName:NSCalibratedRGBColorSpace
							 bytesPerRow:0 
							bitsPerPixel:0
								  toFile:path];
}

/*
- (BOOL)saveTileImageWithPixels:(char **)pixels
					 pixelsWide:(int)width 
					 pixelsHigh:(int)height 
				  bitsPerSample:(int)bps 
				samplesPerPixel:(int)spp 
					   hasAlpha:(BOOL)alpha 
					   isPlanar:(BOOL)isPlanar 
				 colorSpaceName:(NSString *)colorSpaceName 
					bytesPerRow:(int)rowBytes 
				   bitsPerPixel:(int)pixelBits
						 toFile:(NSString *)path

Saves pixels to a TIFF file and opens a window showing the image represented by pixels.
Used for debugging.

 */

- (BOOL)saveTileImageWithPixels:(char **)pixels
					 pixelsWide:(int)width 
					 pixelsHigh:(int)height 
				  bitsPerSample:(int)bps 
				samplesPerPixel:(int)spp 
					   hasAlpha:(BOOL)alpha 
					   isPlanar:(BOOL)isPlanar 
				 colorSpaceName:(NSString *)colorSpaceName 
					bytesPerRow:(int)rowBytes 
				   bitsPerPixel:(int)pixelBits
						 toFile:(NSString *)path

{
	// create an image representation from the pixels and write the rep to file at path
	
	NSBitmapImageRep *tileImageRep = [[NSBitmapImageRep alloc] initWithBitmapDataPlanes:pixels
																			 pixelsWide:width
																			 pixelsHigh:height
																		  bitsPerSample:bps
																		samplesPerPixel:spp
																			   hasAlpha:alpha
																			   isPlanar:isPlanar
																		 colorSpaceName:colorSpaceName
																			bytesPerRow:rowBytes
																		   bitsPerPixel:pixelBits];
	if (!tileImageRep)
	{
		printf("no tileImageRep for %s", [path UTF8String]);
		return NO;
	}
	[[tileImageRep TIFFRepresentation] writeToFile:path atomically:NO];
	
	
	
	if (NO)		// we may need this for debugging - crashes with large images; memory issue?
	{
		// create an image and open a new window with the image
		
		NSImage *tileImage = [[NSImage alloc] initWithSize:NSMakeSize(width, height)];
		if (!tileImage)
		{
			printf("no tileImage for %s", [path UTF8String]);
			return NO;
		}
		[tileImage addRepresentation:tileImageRep];
		
		NSImageView *tileImageView = [[NSImageView alloc] initWithFrame:NSMakeRect(0.0, 0.0, width, height)];
		if (!tileImageView)
		{
			printf("no tileImageView for %s", [path UTF8String]);
			return NO;
		}
		[tileImageView setImage:tileImage];
		
		NSWindow *tileImageWindow = [[NSWindow alloc] initWithContentRect:NSMakeRect(0.0, 0.0, width, height)
																styleMask:NSTitledWindowMask | NSClosableWindowMask | NSMiniaturizableWindowMask
																  backing:NSBackingStoreBuffered
																	defer:YES];
		if (!tileImageWindow)
		{
			printf("no tileImageWindow for %s", [path UTF8String]);
			return NO;
		}
		[tileImageWindow setTitle:path];
		[tileImageWindow setContentView:tileImageView];
		[tileImageWindow orderFront:nil];
	}
	
	
	
	return YES;
}

- (unsigned int)tileIndexForRow:(unsigned int)_row column:(unsigned int)_column
{
	return _row * tileColumns + _column;
}

- (float)imageAspectRatio
{
	return imageAspectRatio;
}

- (BOOL)glPredatesGLOnePointOne
{
	// for now, assume modern OpenGL
	return NO;
}




#pragma mark drawing


- (void)drawRect:(NSRect)rect
{
	// printf("-[%s drawRect:]\n", [[self description] UTF8String]);
	
	
	if (!inited) [self delayedInit];
	
	glClear(GL_COLOR_BUFFER_BIT);
	
	glMatrixMode(GL_PROJECTION);
	
	glLoadIdentity();
	
	GLdouble wOrthoXstart, wOrthoXend, wOrthoYstart, wOrthoYend, wOrthoZstart, wOrthoZend;
	
	if (imageXNeedsFlip)
	{
		wOrthoXstart = orthoXend;
		wOrthoXend = orthoXstart;
	}
	else
	{
		wOrthoXstart = orthoXstart;
		wOrthoXend = orthoXend;
	}
	
	if (imageYNeedsFlip)
	{
		// glOrtho(orthoXstart, orthoXend, orthoYend, orthoYstart, orthoZstart, orthoZend);
		wOrthoYstart = orthoYend;
		wOrthoYend = orthoYstart;
	}
	else
	{
		// glOrtho(orthoXstart, orthoXend, orthoYstart, orthoYend, orthoZstart, orthoZend);
		wOrthoYstart = orthoYstart;
		wOrthoYend = orthoYend;
	}
	
	if (imageZNeedsFlip)
	{
		wOrthoZstart = orthoZend;
		wOrthoZend = orthoZstart;
	}
	else
	{
		wOrthoZstart = orthoZstart;
		wOrthoZend = orthoZend;
	}
	
	glOrtho(wOrthoXstart, wOrthoXend, wOrthoYstart, wOrthoYend, wOrthoZstart, wOrthoZend);
	
	
	
	
	glMatrixMode(GL_MODELVIEW);
	
	
	if (loadSucceeded)
	{
		if (![self inLiveResize])
		{
			if (outputOpenBenchmarks) [self outputBenchmarkWithMessage:@"begin draw tiles\n"];
			
			NSEnumerator *en = [tileArray objectEnumerator];
			Tile *tile;
			while (tile = [en nextObject])
			{
				if ([self shouldDrawTile:tile]) [tile draw];
				if (drawTileOutlines) [tile drawOutline];
			}
			
			if (outputOpenBenchmarks)
			{
				[self outputBenchmarkWithMessage:@"end draw tiles\n"];
				[self outputFileInfo];
				outputOpenBenchmarks = NO;
			}
		}
	}
	else
	{
		glColor4f(0.0, 1.0, 1.0, 1.0);
		glBegin(GL_LINES);
		glVertex2d(0.0, 0.0);
		glVertex2d(1.0, 1.0);
		glVertex2d(0.0, 1.0);
		glVertex2d(1.0, 0.0);
		glEnd();
	}	
	
    [[self openGLContext] flushBuffer];
	//  glutSwapBuffers();
	
	
}


- (BOOL)shouldDrawTile:(Tile *)tile
{
	GLdouble x1, y1, x2, y2, x3, y3, x4, y4;
	GLdouble wOrthoXstart, wOrthoXend, wOrthoYstart, wOrthoYend;
		
	[tile getVerticeX1:&x1
					y1:&y1
					x2:&x2
					y2:&y2
					x3:&x3
					y3:&y3
					x4:&x4
					y4:&y4];
	
	
	if (orthoXend < orthoXstart)  
	{
		wOrthoXstart = orthoXend;
		wOrthoXend = orthoXstart;
	} 
	else
	{
		wOrthoXstart = orthoXstart;
		wOrthoXend = orthoXend;
	}
	
	if (orthoYend < orthoYstart)
	{
		wOrthoYstart = orthoYend;
		wOrthoYend = orthoYstart;
	}
	else
	{
		wOrthoYstart = orthoYstart;
		wOrthoYend = orthoYend;
	}
	
	
	/*
	// if any corner is visible return YES
	
	if ( (wOrthoXstart <= x1 && x1 <= wOrthoXend) && (wOrthoYstart <= y1 && y1 <= wOrthoYend) ) return YES;
	if ( (wOrthoXstart <= x2 && x2 <= wOrthoXend) && (wOrthoYstart <= y2 && y2 <= wOrthoYend) ) return YES;
	if ( (wOrthoXstart <= x3 && x3 <= wOrthoXend) && (wOrthoYstart <= y3 && y3 <= wOrthoYend) ) return YES;
	if ( (wOrthoXstart <= x4 && x4 <= wOrthoXend) && (wOrthoYstart <= y4 && y4 <= wOrthoYend) ) return YES;
	 */
	
	// if the tile is to the left of visible return NO	
	if (x1 < wOrthoXstart && x2 < wOrthoXstart) return NO;
	
	// if the tile is to the right of visible return NO	
	if (x1 > wOrthoXend && x2 > wOrthoXend) return NO;
	
	// if the tile is above visible return NO	
	if (y1 > wOrthoYend && y3 > wOrthoYend) return NO;
	
	// if the tile is below visible return NO
	if (y1 < wOrthoYstart && y3 < wOrthoYstart) return NO;
	
	// a portion of the tile is visible; return YES
	return YES;
}

#pragma mark mouse response




- (void)mouseDragged:(NSEvent *)theEvent
{
	float deltaX = [theEvent deltaX];
	float deltaY = [theEvent deltaY];
	NSPoint deltaVector = NSMakePoint(deltaX, deltaY);
	NSPoint localDeltaVector = [self convertPoint:deltaVector fromView:nil];
	NSRect bounds = [self bounds];
	GLdouble glDeltaX = (GLdouble)localDeltaVector.x / bounds.size.width * orthoDelta;
	GLdouble glDeltaY = (GLdouble)localDeltaVector.y / bounds.size.height * orthoDelta;
	
	if (imageXNeedsFlip)
	{
		orthoXstart += glDeltaX;
	}
	else
	{
		orthoXstart -= glDeltaX;
	}
	orthoXend = orthoXstart + orthoDelta; 
	
	if (imageYNeedsFlip)
	{
		orthoYstart -= glDeltaY;
	}
	else
	{
		orthoYstart += glDeltaY;
	}
	orthoYend = orthoYstart + orthoDelta; 
	
	[self setNeedsDisplay:YES];
	
}



#pragma mark keyboard response




- (void)clearFlags
{
	optionKeyDown =			NO;
	controlKeyDown =		NO;
	commandKeyDown =		NO;
	shiftDown =				NO;
	capsLockDown =			NO;
	
}


- (void)setModifiersFromCarbon
{
	UInt32 carbonOptionKeyMask, carbonRightOptionKeyMask;
	UInt32 carbonShiftKeyMask, carbonRightShiftKeyMask;
	UInt32 carbonControlKeyMask, carbonRightControlKeyMask;
	UInt32 carbonCmdKeyMask;
	UInt32 carbonAlphaLockMask;
	
	carbonOptionKeyMask			= 1 << optionKeyBit;
	carbonRightOptionKeyMask	= 1 << rightOptionKeyBit;
	
	carbonShiftKeyMask			= 1 << shiftKeyBit;
	carbonRightShiftKeyMask		= 1 << rightShiftKeyBit;
	
	carbonControlKeyMask		= 1 << controlKeyBit;
	carbonRightControlKeyMask	= 1 << rightControlKeyBit;
	
	carbonCmdKeyMask			= 1 << cmdKeyBit;
	carbonAlphaLockMask			= 1 << alphaLockBit;
	
	
	UInt32 currentModifiers = GetCurrentEventKeyModifiers();
	
	optionKeyDown =		(carbonOptionKeyMask == (currentModifiers & carbonOptionKeyMask) ) || (carbonRightOptionKeyMask == (currentModifiers & carbonRightOptionKeyMask) );
	shiftDown =			(carbonShiftKeyMask == (currentModifiers & carbonShiftKeyMask) ) || (carbonRightShiftKeyMask == (currentModifiers & carbonRightShiftKeyMask) );
	controlKeyDown =	(carbonControlKeyMask == (currentModifiers & carbonControlKeyMask) ) || (carbonRightControlKeyMask == (currentModifiers & carbonRightControlKeyMask) );
	commandKeyDown =	(carbonCmdKeyMask == (currentModifiers & carbonCmdKeyMask) );
	capsLockDown =		(carbonAlphaLockMask == (currentModifiers & carbonAlphaLockMask) );
}
/*
 */


- (void)setFlagsfromEvent:(NSEvent *)theEvent
{
	unsigned int modifiers = [theEvent modifierFlags];
		
	optionKeyDown =			NSAlternateKeyMask == (modifiers & NSAlternateKeyMask);
	controlKeyDown =		NSControlKeyMask == (modifiers & NSControlKeyMask);
	commandKeyDown =		NSCommandKeyMask == (modifiers & NSCommandKeyMask);
	shiftDown =				NSShiftKeyMask == (modifiers & NSShiftKeyMask);
	capsLockDown =			NSAlphaShiftKeyMask == (modifiers & NSAlphaShiftKeyMask);
}

- (void)flagsChanged:(NSEvent *)theEvent
{
	[self setFlagsfromEvent:theEvent];
}

- (void)keyDown:(NSEvent *)theEvent
{
	NSString *characters = [theEvent characters];
	unichar firstCharacter = [characters characterAtIndex:0];
	
	switch (firstCharacter)
	{
		case '+': 
		case '=': 
			[self zoomIn];
			break;
			
		case '-': 
			[self zoomOut];
			break;
			
		case 'o': 
			[self toggleDrawTileOutlines];
			break;
			
		case NSLeftArrowFunctionKey:
			[self translateForLeftArrowDown];
			break;
			
		case NSRightArrowFunctionKey:
			[self translateForRightArrowDown];
			break;
			
		case NSUpArrowFunctionKey:
			[self translateForUpArrowDown];
			break;
			
		case NSDownArrowFunctionKey:
			[self translateForDownArrowDown];
			break;
			
	}
}


- (BOOL)acceptsFirstResponder
{
	return YES;
}



#pragma mark zoom and translation




- (void)zoomIn
{
	if (orthoDelta > 0.00001)				// limit zoom in
	{
		GLdouble controlFactor = 1.0;
		
		if (optionKeyDown) controlFactor = 2.0;
		if (controlKeyDown) controlFactor = 10.0;
		
		oldOrthoDelta = orthoDelta;
		orthoDelta *= pow(orthoDeltaScaleFactor, controlFactor);
		
		[self finishZoom];
		[self setNeedsDisplay:YES];
	}
}


- (void)zoomOut
{
	GLdouble controlFactor = 1.0;
	
	if (optionKeyDown) controlFactor = 2.0;
	if (controlKeyDown) controlFactor = 10.0;
	
	GLdouble workingOrthoDelta = orthoDelta / pow(orthoDeltaScaleFactor, controlFactor);
	
	if (workingOrthoDelta <= 1.0)		// zoom out only to 1:1 - avoids "greyout blues" with very large images
	{
		oldOrthoDelta = orthoDelta;
		orthoDelta = workingOrthoDelta;
		
		[self finishZoom];
		[self setNeedsDisplay:YES];
	}
}




- (void)finishZoom
{
	orthoXstart -= 0.5 * (orthoDelta - oldOrthoDelta);
	orthoYstart -= 0.5 * (orthoDelta - oldOrthoDelta);
	
	orthoXend = orthoXstart + orthoDelta; 
	orthoYend = orthoYstart + orthoDelta;
}

- (void)translateForLeftArrowDown
{
	if (controlKeyDown) {[self translateImageLeftEdgeToViewLeftEdge]; return;}
	if (optionKeyDown) {[self translateLeftByFraction:0.2];} else {[self translateLeftByFraction:0.1];}
	[self setNeedsDisplay:YES];
}

- (void)translateForRightArrowDown
{
	if (controlKeyDown) {[self translateImageRightEdgeToViewRighgtEdge]; return;}
	if (optionKeyDown) {[self translateRightByFraction:0.2];} else {[self translateRightByFraction:0.1];}
	[self setNeedsDisplay:YES];
}

- (void)translateForUpArrowDown
{
	if (controlKeyDown) {[self translateImageTopEdgeToViewTopEdge]; return;}
	if (optionKeyDown) {[self translateUpByFraction:0.2];} else {[self translateUpByFraction:0.1];}
	[self setNeedsDisplay:YES];
}

- (void)translateForDownArrowDown
{
	if (controlKeyDown) {[self translateImageBottomEdgeToViewBottomEdge]; return;}
	if (optionKeyDown) {[self translateDownByFraction:0.2];} else {[self translateDownByFraction:0.1];}
	[self setNeedsDisplay:YES];
}

- (void)translateRightByFraction:(GLdouble)fraction
{
	if (imageXNeedsFlip)
	{
		orthoXstart -= orthoDelta * fraction;
		orthoXend = orthoXstart + orthoDelta;
	}
	else
	{
		orthoXstart += orthoDelta * fraction;
		orthoXend = orthoXstart + orthoDelta;
	}
}

- (void)translateLeftByFraction:(GLdouble)fraction
{
	if (imageXNeedsFlip)
	{
		orthoXstart += orthoDelta * fraction;
		orthoXend = orthoXstart + orthoDelta;
	}
	else
	{
		orthoXstart -= orthoDelta * fraction;
		orthoXend = orthoXstart + orthoDelta;
	}
}

- (void)translateUpByFraction:(GLdouble)fraction
{
	if (imageYNeedsFlip)
	{
		orthoYstart -= orthoDelta * fraction;
	}
	else
	{
		orthoYstart += orthoDelta * fraction;
	}
	orthoYend = orthoYstart + orthoDelta; 
}

- (void)translateDownByFraction:(GLdouble)fraction
{
	if (imageYNeedsFlip)
	{
		orthoYstart += orthoDelta * fraction;
	}
	else
	{
		orthoYstart -= orthoDelta * fraction;
	}
	orthoYend = orthoYstart + orthoDelta; 
}

- (void)translateImageLeftEdgeToViewLeftEdge
{
	if (imageXNeedsFlip)
	{
	orthoXend = 1.0;
	orthoXstart = orthoXend - orthoDelta;
		
	}
	else
	{
	orthoXstart = 0.0;
	orthoXend = orthoXstart + orthoDelta;
		
	}
	// orthoXstart = 0.0;
	// orthoXend = orthoXstart + orthoDelta;
}

- (void)translateImageRightEdgeToViewRighgtEdge
{
	if (imageXNeedsFlip)
	{
	orthoXstart = 0.0;
	orthoXend = orthoXstart + orthoDelta;
		
	}
	else
	{
	orthoXend = 1.0;
	orthoXstart = orthoXend - orthoDelta;
		
	}
	// orthoXend = 1.0;
	// orthoXstart = orthoXend - orthoDelta;
}

- (void)translateImageTopEdgeToViewTopEdge
{
	if (imageYNeedsFlip)
	{
		orthoYstart = 0.0;
		orthoYend = orthoYstart + orthoDelta; 
	}
	else
	{
		orthoYend = 1.0;
		orthoYstart = orthoYend - orthoDelta;
	}
}

- (void)translateImageBottomEdgeToViewBottomEdge
{
	if (imageYNeedsFlip)
	{
		orthoYend = 1.0;
		orthoYstart = orthoYend - orthoDelta;
	}
	else
	{
		orthoYstart = 0.0;
		orthoYend = orthoYstart + orthoDelta; 
	}
}



#pragma mark diagnostic




- (void)toggleDrawTileOutlines
{
	drawTileOutlines = !drawTileOutlines;
	[self setNeedsDisplay:YES];
}

- (void)startBenchmark
{
	Microseconds(&startBenchmarkUW);
}

- (void)outputBenchmarkWithMessage:(NSString *)message
{
	UnsignedWide currentBenchmarkUW, deltaBenchmarkUW;
	Microseconds(&currentBenchmarkUW);
	
	deltaBenchmarkUW.lo = currentBenchmarkUW.lo - startBenchmarkUW.lo;
	[self logMessageToLIVLog:[NSString stringWithFormat:@"T + \t%u\t microseconds: %@", deltaBenchmarkUW.lo, message]];
}

- (void)outputFileInfo
{
	// assumes imagePixelsWide, imagePixelsHigh, fileNameForDisplay already set
	// call after image is loaded
	
	[self logMessageToLIVLog:fileNameForDisplay];
	[self logMessageToLIVLog:nlStr];
	[self logMessageToLIVLog:[NSString stringWithFormat:@"%d\t x \t%d\t pixels\n", imagePixelsWide, imagePixelsHigh]];
	
	
	NSFileManager *fileManager = [NSFileManager defaultManager];
	NSDictionary *attr = [fileManager fileAttributesAtPath:filePath traverseLink:YES];
	if (attr)
	{
		NSNumber *fileSize = [attr objectForKey:NSFileSize];
		if (fileSize)
		{
			[self logMessageToLIVLog:[NSString stringWithFormat:@"%d\t bytes\n", [fileSize unsignedLongValue]]];
		}
	}
	[self logMessageToLIVLog:nlStr];
	[self logMessageToLIVLog:nlStr];
}

- (NSString *)diagnosticFolderPathForComponent:(NSString *)_component
{
	NSString *_diagnosticFolderPath, *docSubdirectory;
	// BOOL isDir;
	NSDictionary *pathAttributes;
	// NSError *err;
	
	NSFileManager *fileManager = [NSFileManager defaultManager];
	
	// NSString *documentsFolderPath = [@"~/Documents/" stringByExpandingTildeInPath];
	
	// NSRange range = NSMakeRange(0, 1);
	NSString *compFirst = [_component substringToIndex:1];
	NSString *compFirstLower = [compFirst lowercaseString];
	NSString *compFirstUpper = [compFirst uppercaseString];
	
	// comments below assume _component is  @"LIV Debug Files"; this folder would go into 
	// "~/Documents/L" or "~/Documents/l" if one of these folders exist. Otherwise, the folder will go in "~/Documents"
	
	NSString *documentsFolderPath = [NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES) objectAtIndex:0];
	
	docSubdirectory = [documentsFolderPath stringByAppendingPathComponent:compFirstUpper];
	if ( !isDirectoryAtPath(docSubdirectory) )					// if "~/Documents/L" exists, put the debug folder in it
	{
		docSubdirectory = [documentsFolderPath stringByAppendingPathComponent:compFirstLower];
		if 	( !isDirectoryAtPath(docSubdirectory) )				// if "~/Documents/l" exists, put the debug folder in it
		{
			docSubdirectory = documentsFolderPath;				// no "~/Documents/L" or "~/Documents/l" exists, put the debug folder in "~/Documents/"
		}
	}
	
	_diagnosticFolderPath = [docSubdirectory stringByAppendingPathComponent:_component];		// @"LIV Debug Files"
	
	if (!isDirectoryAtPath(_diagnosticFolderPath))
	{
		if ([fileManager fileExistsAtPath:_diagnosticFolderPath])
		{
			int pathCount = 1;
			NSString *formatDebugFolderPath = [_diagnosticFolderPath stringByAppendingString:@"%d"];
			NSString *workingDebugFolderPath = [NSString stringWithFormat:formatDebugFolderPath, pathCount];
			
			while ([fileManager fileExistsAtPath:workingDebugFolderPath]) 
			{
				pathCount++;
				workingDebugFolderPath = [NSString stringWithFormat:formatDebugFolderPath, pathCount];
			}
			_diagnosticFolderPath = workingDebugFolderPath;
		}
		
		pathAttributes = [fileManager fileAttributesAtPath:documentsFolderPath traverseLink:NO]; 
		[fileManager createDirectoryAtPath:_diagnosticFolderPath attributes:pathAttributes];
	}
	return _diagnosticFolderPath;
}


- (NSString *)folderForTIFFsFromFile:(NSString *)_fileNameForDisplay
{
	NSString *folderForTIFFsFromFile;
	
	int length = [_fileNameForDisplay length];
	int prefixLength = (length < 10) ? length : 10;
	NSString *prefix = [_fileNameForDisplay substringToIndex:prefixLength];
	folderForTIFFsFromFile = [[[diagnosticFolderPath stringByAppendingPathComponent:prefix] stringByAppendingString:@" "] stringByAppendingString:[self dateTimeStampAsString]];
	NSFileManager *fileManager = [NSFileManager defaultManager];
	if (!isDirectoryAtPath(folderForTIFFsFromFile)) 
	{
		[fileManager createDirectoryAtPath:folderForTIFFsFromFile attributes:nil];
	}
	return folderForTIFFsFromFile;
}


/*
 - (NSString *)dateTimeStampAsString
{
	NSDate *date = [NSDate date];
	NSDateFormatter *dateFormatter = [[[NSDateFormatter alloc] init] autorelease];
	[dateFormatter setFormatterBehavior:NSDateFormatterBehavior10_4];
	[dateFormatter setDateFormat:@"yyyy-MM-dd HH-mm-ss"];
	return [dateFormatter stringFromDate:date];
}
*/


- (NSString *)dateTimeStampAsString
{
	NSDate *date = [NSDate date];
	return [[self dateFormatter] stringFromDate:date];
}


- (NSDateFormatter *)dateFormatter
{
	if (!dateFormatter)
	{
		dateFormatter = [[NSDateFormatter alloc] init];
		[dateFormatter setFormatterBehavior:NSDateFormatterBehavior10_4];
		[dateFormatter setDateFormat:@"yyyy-MM-dd HH-mm-ss"];
	}
	return dateFormatter;
}


- (NSString *)determineLIVLogFlePath
{
	NSString *_livLogFilePath = [diagnosticFolderPath stringByAppendingPathComponent:livLog];
	return _livLogFilePath;
}


- (void)logMessageToLIVLog:(NSString *)msg
{
	FILE * file = fopen([livLogFilePath UTF8String], "a");
	// fprintf(file, [[msg stringByAppendingString:nlStr] UTF8String]);
	fprintf(file, [msg UTF8String]);
	fclose(file);
}


- (void)dateTimeStampToLog
{
	[self logMessageToLIVLog:[self dateTimeStampAsString]];
	[self logMessageToLIVLog:nlStr];
}



#pragma mark accessors



- (NSString *)colorspaceName
{
	return colorspaceName;
}


- (void)setColorspaceName:(NSString *)_colorspaceName
{
	[_colorspaceName retain];
	[colorspaceName release];
	colorspaceName = _colorspaceName;
}


- (NSMutableArray *)tileArray
{
	return tileArray;
}


- (void)setTileArray:(NSMutableArray *)_tileArray
{
	[_tileArray retain];
	[tileArray release];
	tileArray = _tileArray;
}


- (NSString *)fileNameForDisplay
{
	return fileNameForDisplay;
}


- (void)setFileNameForDisplay:(NSString *)_fileNameForDisplay
{
	[_fileNameForDisplay retain];
	[fileNameForDisplay release];
	fileNameForDisplay = _fileNameForDisplay;
}


- (NSString *)filePath
{
	return filePath;
}


- (void)setFilePath:(NSString *)_filePath
{
	[_filePath retain];
	[filePath release];
	filePath = _filePath;
}


- (NSString *)diagnosticFolderPath
{
	return diagnosticFolderPath;
}


- (void)setDiagnosticFolderPath:(NSString *)_diagnosticFolderPath
{
	[_diagnosticFolderPath retain];
	[diagnosticFolderPath release];
	diagnosticFolderPath = _diagnosticFolderPath;
}


- (NSString *)livLogFilePath
{
	return livLogFilePath;
}


- (void)setLivLogFilePath:(NSString *)_livLogFilePath
{
	[_livLogFilePath retain];
	[livLogFilePath release];
	livLogFilePath = _livLogFilePath;
}




@end
