//
//  Tile.m
//  Large Image Viewer
//
//  Created by Tom Bernard on 10/19/06.
//  Copyright 2006 Bersearch Information Services. All rights reserved.
//

#import "Tile.h"

void printTileGLError(char *label, GLenum error)
{
	char *errStr;
	
	switch (error)
	{
		case GL_NO_ERROR:
			errStr = "GL_NO_ERROR";
			break;
			
		case GL_INVALID_ENUM:
			errStr = "GL_INVALID_ENUM";
			break;
			
		case GL_INVALID_VALUE:
			errStr = "GL_INVALID_VALUE";
			break;
			
		case GL_INVALID_OPERATION:
			errStr = "GL_INVALID_OPERATION";
			break;
			
		case GL_STACK_OVERFLOW:
			errStr = "GL_STACK_OVERFLOW";
			break;
			
		case GL_STACK_UNDERFLOW:
			errStr = "GL_STACK_UNDERFLOW";
			break;
			
		case GL_OUT_OF_MEMORY:
			errStr = "GL_OUT_OF_MEMORY";
			break;
			
		default: errStr = "undefined error";
	}
	printf("%s%s\n", label, errStr);
}

GLenum pixelTypeForBPS(unsigned int bitsPerSample)
{
	GLenum pixelTypeForBPS = GL_UNSIGNED_BYTE;		// default to GL_UNSIGNED_BYTE
	switch(bitsPerSample)
	{
		// case 8:
			// pixelTypeForBPS =  GL_UNSIGNED_BYTE;
		
		case 16:
			pixelTypeForBPS = GL_UNSIGNED_SHORT;
			break;
			
		case 32:
			pixelTypeForBPS = GL_UNSIGNED_INT;
			break;
	}
	return pixelTypeForBPS; 
}


GLint internalFormatForColorspace(NSString *_colorSpaceName, BOOL _imageHasAlpha)
{
	if ([_colorSpaceName isEqualToString:NSCalibratedRGBColorSpace])		return _imageHasAlpha ? GL_RGBA : GL_RGB;
	if ([_colorSpaceName isEqualToString:NSDeviceRGBColorSpace])			return _imageHasAlpha ? GL_RGBA : GL_RGB;
	
	if ([_colorSpaceName isEqualToString:NSCalibratedWhiteColorSpace])		return _imageHasAlpha ? GL_LUMINANCE_ALPHA : GL_LUMINANCE;
	if ([_colorSpaceName isEqualToString:NSDeviceWhiteColorSpace])			return _imageHasAlpha ? GL_LUMINANCE_ALPHA : GL_LUMINANCE;
	
	// if ([_colorSpaceName isEqualToString:NSCalibratedBlackColorSpace])	return _imageHasAlpha ? : ;
	// if ([_colorSpaceName isEqualToString:NSDeviceBlackColorSpace])		return _imageHasAlpha ? : ;
	
	// if ([_colorSpaceName isEqualToString:NSDeviceCMYKColorSpace])		return _imageHasAlpha ? : ;
	// if ([_colorSpaceName isEqualToString:NSNamedColorSpace])				return _imageHasAlpha ? : ;
	// if ([_colorSpaceName isEqualToString:NSPatternColorSpace])			return _imageHasAlpha ? : ;
	// if ([_colorSpaceName isEqualToString:NSCustomColorSpace])			return _imageHasAlpha ? : ;
	
	return GL_RGBA;
}

// 	if ([_colorSpaceName isEqualToString:]) return _imageHasAlpha ? : ;


GLint formatForColorspace(NSString *_colorSpaceName, BOOL _imageHasAlpha)
{
	if ([_colorSpaceName isEqualToString:NSCalibratedRGBColorSpace])		return _imageHasAlpha ? GL_RGBA : GL_RGB;
	if ([_colorSpaceName isEqualToString:NSDeviceRGBColorSpace])			return _imageHasAlpha ? GL_RGBA : GL_RGB;
	
	if ([_colorSpaceName isEqualToString:NSCalibratedWhiteColorSpace])		return _imageHasAlpha ? GL_LUMINANCE_ALPHA : GL_LUMINANCE;
	if ([_colorSpaceName isEqualToString:NSDeviceWhiteColorSpace])			return _imageHasAlpha ? GL_LUMINANCE_ALPHA : GL_LUMINANCE;
	
	// if ([_colorSpaceName isEqualToString:NSCalibratedBlackColorSpace])	return _imageHasAlpha ? : ;
	// if ([_colorSpaceName isEqualToString:NSDeviceBlackColorSpace])		return _imageHasAlpha ? : ;
	
	// if ([_colorSpaceName isEqualToString:NSDeviceCMYKColorSpace])		return _imageHasAlpha ? : ;
	// if ([_colorSpaceName isEqualToString:NSNamedColorSpace])				return _imageHasAlpha ? : ;
	// if ([_colorSpaceName isEqualToString:NSPatternColorSpace])			return _imageHasAlpha ? : ;
	// if ([_colorSpaceName isEqualToString:NSCustomColorSpace])			return _imageHasAlpha ? : ;

	return GL_RGBA;
}


@implementation Tile

/*
+ (unsigned char *)allocateTextureBitmapDataRows:(int)tpRows colums:(int)tpColumns samplesPerPixel:(int)spp bitsPerSample:(int)bps
{
	return calloc(tpRows * tpColumns * spp * bps, 1);
}
 */

- (id)initWithTexelWidth:(GLuint)_width 
			 texelHeight:(GLuint)_height 
					 row:(GLuint)_row
				  column:(GLuint)_column
		   tileNeedsFlip:(BOOL)_needsFlip
		   colorspaeName:(NSString *)_colorspaceName
		   imageHasAlpha:(BOOL)_imageHasAlpha
		   bitsPerSample:(unsigned int)_bitsPerSample
				  pixels:(const GLvoid *)_pixels;
{
	if (self = [super init])
	{
		// tileTexture = _texture;
		tileTexelWidth = _width;
		tileTexelHeight = _height;
		tileRow = _row;
		tileColumn = _column;
		tileNeedsFlip = _needsFlip;
		[self setColorspaceName:_colorspaceName];
		imageHasAlpha = _imageHasAlpha;
		bitsPerSample = _bitsPerSample;
		
		// glEnable (GL_TEXTURE_2D);							// already enabled
		// glPixelStorei(GL_UNPACK_ROW_LENGTH, textureWidth);		// not needed?
		// glPixelStorei(GL_UNPACK_ALIGNMENT, 1);					// not needed?
		
		GLenum glErr = glGetError();
		if (glErr)
		{
			printTileGLError("Tile init failed at start with ", glErr);
			[self release];
			return nil;
		}
		
		if ([self glPredatesGLOnePointOne])
		{
			printf("glGenTextures requires OpenGL 1.1 or later. Cannot open image.\n");
			[self release];
			return nil;
		}
		
		glGenTextures(1, &tileTexture);
		
		if (!tileTexture)
		{
			glErr = glGetError();
			printTileGLError("No texture generated. ", glErr);
			[self release];
			return nil;
			
		}
		
		glBindTexture (GL_TEXTURE_2D, tileTexture);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		
		glErr = glGetError();
		if (glErr)
		{
			printTileGLError("Texture setup failed with ", glErr);
			[self release];
			return nil;
		}
		
		/*
		glTexImage2D(GL_TEXTURE_2D,
					 0, 
					 imageHasAlpha ? GL_RGBA : GL_RGB, 
					 tileTexelWidth,
					 tileTexelHeight,
					 0,
					 imageHasAlpha ? GL_RGBA : GL_RGB, 
					 pixelTypeForBPS(bitsPerSample),			// GL_UNSIGNED_BYTE	// GL_UNSIGNED_SHORT	// GL_UNSIGNED_INT
					 _pixels);
		 */
		
		glTexImage2D(GL_TEXTURE_2D,
					 0, 
					 internalFormatForColorspace(colorspaceName, imageHasAlpha), 
					 tileTexelWidth,
					 tileTexelHeight,
					 0,
					 formatForColorspace(colorspaceName, imageHasAlpha), 
					 pixelTypeForBPS(bitsPerSample),			// GL_UNSIGNED_BYTE	// GL_UNSIGNED_SHORT	// GL_UNSIGNED_INT
					 _pixels);
		
		glErr = glGetError();
		GLboolean ts = glIsTexture(tileTexture);
		if (glErr || !ts)
		{
			printTileGLError("glTexImage2D failed with ", glErr);
			printf("ts = %s\n", ts ? "true" : "false");
			[self release];
			return nil;
		}
	}
	return self;
}

- (void)dealloc
{
	[self deleteTexture];
	[colorspaceName release];
	[super dealloc];
}


- (void)setVerticeX1:(GLdouble)_x1
				  y1:(GLdouble)_y1
				  x2:(GLdouble)_x2
				  y2:(GLdouble)_y2
				  x3:(GLdouble)_x3
				  y3:(GLdouble)_y3
				  x4:(GLdouble)_x4
				  y4:(GLdouble)_y4
{
	x1 = _x1;
	y1 = _y1;
	x2 = _x2;
	y2 = _y2;
	x3 = _x3;
	y3 = _y3;
	x4 = _x4;
	y4 = _y4;
	
}

- (void)getVerticeX1:(GLdouble *)_x1
				  y1:(GLdouble *)_y1
				  x2:(GLdouble *)_x2
				  y2:(GLdouble *)_y2
				  x3:(GLdouble *)_x3
				  y3:(GLdouble *)_y3
				  x4:(GLdouble *)_x4
				  y4:(GLdouble *)_y4;
{
	*_x1 = x1;
	*_y1 = y1;
	*_x2 = x2;
	*_y2 = y2;
	*_x3 = x3;
	*_y3 = y3;
	*_x4 = x4;
	*_y4 = y4;
}


- (GLdouble)x1 {return x1;}
- (GLdouble)y1 {return y1;}

- (GLdouble)x2 {return x2;}
- (GLdouble)y2 {return y2;}

- (GLdouble)x3 {return x3;}
- (GLdouble)y3 {return y3;}

- (GLdouble)x4 {return x4;}
- (GLdouble)y4 {return y4;}

- (void)setTextureCoordinateS1:(GLdouble)_s1
							t1:(GLdouble)_t1
							s2:(GLdouble)_s2
							t2:(GLdouble)_t2
							s3:(GLdouble)_s3
							t3:(GLdouble)_t3
							s4:(GLdouble)_s4
							t4:(GLdouble)_t4
{
	s1 = _s1;
	t1 = _t1;
	s2 = _s2;
	t2 = _t2;
	s3 = _s3;
	t3 = _t3;
	s4 = _s4;
	t4 = _t4;
}


- (void)draw
{
	glBindTexture (GL_TEXTURE_2D, tileTexture);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);		//  // GL_NEAREST // GL_NEAREST_MIPMAP_NEAREST // GL_NEAREST_MIPMAP_LINEAR
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);	// GL_CLAMP	// GL_CLAMP_TO_BORDER // GL_CLAMP_TO_EDGE
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);	// GL_CLAMP	// GL_CLAMP_TO_BORDER // GL_CLAMP_TO_EDGE
	
	glBegin(GL_QUADS);
	glColor4f(1.0, 1.0, 1.0, 1.0);
	
	if (tileNeedsFlip) {glTexCoord2d(s4, t4);} else {glTexCoord2d(s1, t1);}
	glVertex2d(x1, y1);
	
	if (tileNeedsFlip) {glTexCoord2d(s3, t3);} else {glTexCoord2d(s2, t2);}
	glVertex2d(x2, y2);
	
	if (tileNeedsFlip) {glTexCoord2d(s2, t2);} else {glTexCoord2d(s3, t3);}
	glVertex2d(x3, y3);
	
	if (tileNeedsFlip) {glTexCoord2d(s1, t1);} else {glTexCoord2d(s4, t4);}
	glVertex2d(x4, y4);
	
	glEnd();
	
	// printf("tileTexture = %d\n", tileTexture);
}

- (void)drawOutline
{
	glColor4f(0.0, 1.0, 0.0, 1.0);		//  green
	
	glBegin(GL_LINE_LOOP);
	
	glVertex2d(x1, y1);
	glVertex2d(x2, y2);
	glVertex2d(x3, y3);
	glVertex2d(x4, y4);
	
	glEnd();
}

- (void)deleteTexture
{
	glDeleteTextures(1, &tileTexture);
}

- (BOOL)glPredatesGLOnePointOne
{
	// for now, assume modern OpenGL
	return NO;
}



- (NSString *)colorspaceName
{
	return colorspaceName;
}


- (void)setColorspaceName:(NSString *)_colorspaceName
{
	[_colorspaceName retain];
	[colorspaceName release];
	colorspaceName = _colorspaceName;
}


/*
 */


/*
- (id)retain
{
	return [super retain];
}

- (oneway void)release
{
	[super release];
}
 */



@end
